@@ -17,6 +17,10 @@ from botocore import xform_name
 
 from .params import create_request_parameters
 from .response import RawHandler, ResourceHandler
+from .model import Action
+
+from boto3.docs.docstring import ActionDocstring
+from boto3.utils import inject_attribute
 
 
 logger = logging.getLogger(__name__)
@@ -31,25 +35,27 @@ class ServiceAction(object):
 
     :type action_model: :py:class`~boto3.resources.model.Action`
     :param action_model: The action model.
+
     :type factory: ResourceFactory
     :param factory: The factory that created the resource class to which
                     this action is attached.
-    :type resource_defs: dict
-    :param resource_defs: Service resource definitions.
-    :type service_model: :ref:`botocore.model.ServiceModel`
-    :param service_model: The Botocore service model
+
+    :type service_context: :py:class:`~boto3.utils.ServiceContext`
+    :param service_context: Context about the AWS service
     """
-    def __init__(self, action_model, factory=None, resource_defs=None,
-                 service_model=None):
+    def __init__(self, action_model, factory=None, service_context=None):
         self._action_model = action_model
 
         # In the simplest case we just return the response, but if a
         # resource is defined, then we must create these before returning.
-        resource = action_model.resource
-        if resource:
-            self._response_handler = ResourceHandler(resource.path,
-                factory, resource_defs, service_model, resource,
-                action_model.request.operation)
+        resource_response_model = action_model.resource
+        if resource_response_model:
+            self._response_handler = ResourceHandler(
+                search_path=resource_response_model.path,
+                factory=factory, resource_model=resource_response_model,
+                service_context=service_context,
+                operation_name=action_model.request.operation
+            )
         else:
             self._response_handler = RawHandler(action_model.path)
 
@@ -71,10 +77,10 @@ class ServiceAction(object):
         params = create_request_parameters(parent, self._action_model.request)
         params.update(kwargs)
 
-        logger.info('Calling %s:%s with %r', parent.meta['service_name'],
+        logger.info('Calling %s:%s with %r', parent.meta.service_name,
             operation_name, params)
 
-        response = getattr(parent.meta['client'], operation_name)(**params)
+        response = getattr(parent.meta.client, operation_name)(**params)
 
         logger.debug('Response: %r', response)
 
@@ -89,15 +95,15 @@ class BatchAction(ServiceAction):
     S3 objects in a single operation rather than calling ``.delete()`` on
     each one individually.
 
-    :type action_model: :py:class:`~boto3.resources.model.Action`
+    :type action_model: :py:class`~boto3.resources.model.Action`
     :param action_model: The action model.
+
     :type factory: ResourceFactory
     :param factory: The factory that created the resource class to which
                     this action is attached.
-    :type resource_defs: dict
-    :param resource_defs: Service resource definitions.
-    :type service_model: :ref:`botocore.model.ServiceModel`
-    :param service_model: The Botocore service model
+
+    :type service_context: :py:class:`~boto3.utils.ServiceContext`
+    :param service_context: Context about the AWS service
     """
     def __call__(self, parent, *args, **kwargs):
         """
@@ -120,17 +126,18 @@ class BatchAction(ServiceAction):
         # the necessary parameters and call the batch operation.
         for page in parent.pages():
             params = {}
-            for resource in page:
+            for index, resource in enumerate(page):
                 # There is no public interface to get a service name
                 # or low-level client from a collection, so we get
                 # these from the first resource in the collection.
                 if service_name is None:
-                    service_name = resource.meta['service_name']
+                    service_name = resource.meta.service_name
                 if client is None:
-                    client = resource.meta['client']
+                    client = resource.meta.client
 
                 create_request_parameters(
-                    resource, self._action_model.request, params=params)
+                    resource, self._action_model.request,
+                    params=params, index=index)
 
             if not params:
                 # There are no items, no need to make a call.
@@ -186,11 +193,50 @@ class WaiterAction(object):
         params.update(kwargs)
 
         logger.info('Calling %s:%s with %r',
-                    parent.meta['service_name'],
+                    parent.meta.service_name,
                     self._waiter_resource_name, params)
 
-        client = parent.meta['client']
+        client = parent.meta.client
         waiter = client.get_waiter(client_waiter_name)
         response = waiter.wait(**params)
 
         logger.debug('Response: %r', response)
+
+
+class CustomModeledAction(object):
+    """A custom, modeled action to inject into a resource."""
+    def __init__(self, action_name, action_model,
+                 function, event_emitter):
+        """
+        :type action_name: str
+        :param action_name: The name of the action to inject, e.g. 'delete_tags'
+
+        :type action_model: dict
+        :param action_model: A JSON definition of the action, as if it were
+            part of the resource model.
+
+        :type function: function
+        :param function: The function to perform when the action is called.
+            The first argument should be 'self', which will be the resource
+            the function is to be called on.
+
+        :type event_emitter: :py:class:`botocore.hooks.BaseEventHooks`
+        :param event_emitter: The session event emitter.
+        """
+        self.name = action_name
+        self.model = action_model
+        self.function = function
+        self.emitter = event_emitter
+
+    def inject(self, class_attributes, service_context, event_name, **kwargs):
+        resource_name = event_name.rsplit(".")[-1]
+        action = Action(self.name, self.model, {})
+        self.function.__name__ = self.name
+        self.function.__doc__ = ActionDocstring(
+            resource_name=resource_name,
+            event_emitter=self.emitter,
+            action_model=action,
+            service_model=service_context.service_model,
+            include_signature=False
+        )
+        inject_attribute(class_attributes, self.name, self.function)
